"""
Author Aurore Guillevic
email  aurore.guillevic@inria.fr
date   May 7, 2019

Example of NFS with a complex quadratic number field f=x^2+5 for p=1109.
"""
# import printing functions for latex output
from examples_pretty_print import pretty_print_product, pretty_print_tex_product, pretty_print_units

p = 1109     # prime, p = 1 mod 4
r = (p-1)//4 # prime order of the subgroup
Fr = GF(r)   # for the linear algebra
Fp = GF(p)
g_ = Fp(2) # generator
g4 = g_^((p-1)//4) # of order 4

ZZx.<x> = ZZ[]

print("p = {} is prime".format(p))
print("(p-1)/4 = {} is prime".format(r))

print("Polynomial selection: complex quadratic")
f = x^2+5
# 33^2 +5*2^2 = 1109
g = 2*x - 33
m = Fp(33/2) # common root mod p

Fpz.<z> = Fp[] # polynomial ring modulo p
assert gcd(Fpz(f), Fpz(g)) == Fpz(x-m)

assert f(m) == 0
assert g(m) == 0

assert f.is_irreducible()
assert (f(m) % p) == 0
assert (f.resultant(g) % p) == 0

print("m = {} # root of f and g mod p\nf = {}\ng = {}".format(m,f,g))

fp = Fpz(f)    # polynomial f modulo p
print("f = {} mod p".format(fp.factor()))
Kf.<af> = NumberField(f)
Clf = Kf.class_number()
print("Cl(Kf) = {}".format(Clf))
print("units(Kf) = {}".format(Kf.units()))
print("number of real roots, pairs of complex roots: {}".format(Kf.signature()))

print("resultant(f,g) = {} = {} mod p".format(f.resultant(g), f.resultant(g) % p))
disc_f =  f.discriminant()
print("disc(f) = {} = {}\n".format(disc_f, disc_f.factor()))

A = 11 # a,b in the relation collection step: a in [-A,A], b in [1,A]

Bf=13
Bg=13
print("chosen smoothness bounds: f-side {}, g-side {}".format(Bf,Bg))

# a dictionary for the factor basis on each side
Ff = dict()
Fg = dict()
print("compute the factor basis on f-side: factor f mod l for each prime 2 <= l <= Bf={}".format(Bf))
for l in prime_range(2, Bf+1):
    Fl = GF(l)     # prime finite field of cardinality l
    Flz.<z> = Fl[] # polynomial ring over Fl
    fl = Flz(f)    # f mod l
    if len(fl.roots())>0: # f has root(s) mod l, so there are as many distinct degree 1 prime ideals above l as distinct roots
        Ff[l] = [ZZx(fi) for (fi,ei) in fl.factor() if fi.degree() == 1]
    print("{}: {}".format(l, (fl).factor()))

Fg = [l for l in prime_range(Bg+1)]
print("\nfactor basis g: {}".format(Fg))

Kf.<af> = NumberField(f)
Clf = Kf.class_number()
print("Class number(Kf) = {}".format(Clf))

F_f = dict() # a second dict with generators of the ideals
for (l,li) in Ff.items():
    for d in li:
        Il = Kf.ideal([l,d])
        gen_Il = Il.gens_reduced()
        if len(gen_Il) == 1: # means it is principal and one generator was found
            if F_f.has_key(l):
                F_f[l].append(gen_Il[0])
            else:
                F_f[l] = [gen_Il[0]]

print("ideals above primes on f-side:")
for l in prime_range(Bf+1):
    if Ff.has_key(l) and F_f.has_key(l) and len(Ff[l]) == len(F_f[l]):
        for idx in range(len(Ff[l])):
            print("({:2d},{:6}), generator {}".format(l, Ff[l][idx], F_f[l][idx]))
    elif Ff.has_key(l):
        for id_l in Ff[l]:
            print("({:2d},{:6}), generator: no".format(l, id_l))

# the ideals are not principal,
# but there is no unit here, so no need for generators

# define two dictionaries to map each prime of Fg, and each ideal of Ff, to a column index in the matrix,
# in order to translate each relation into a row in the matrix
# since g is rational, put Fg first in the matrix, starting with 2, so that the logarithms will be in basis 2
idx_ideal_g = {Fg[idx]:idx for idx in range(len(Fg))}
idx = len(Fg)
lc_g = g.leading_coefficient()
if lc_g != 1:
    idx_inv_V = idx # for 1/V = 1/2 because g is not monic
    idx += 1
idx_ideal_f = dict()

#for (l,li) in Ff.items(): it does not provide the ideals by increasing size
for l in prime_range(Bf+1):
    if Ff.has_key(l):
        for id_li in Ff[l]:
            idx_ideal_f[(l,id_li)] = idx
            idx += 1

print("\nthe correspondance between ideals and column indices is the folowing")
print idx_ideal_g
print idx_ideal_f

row_id = 0
rows = idx
cols = idx
print("{} columns, {} relations meeded".format(cols,rows))

# put 1 for the coefficient 1/V of g
mat = [[0 for j in range(cols)] for i in range(rows)]
if lc_g != 1:
    for ri in range(rows):
        mat[ri][idx_inv_V] = 1

print("start relation collection, a in [-{0},{0}], b in [1,{0}]\n".format(A))

for a in range(-A,A+1):
    for b in [b for b in range(1,A+1) if gcd(a,b) == 1]:
        nf = ZZ(f.resultant(a-b*x))
        ng = ZZ(g.resultant(a-b*x))

        if nf == 0 or ng == 0:
            continue
        fac_f = nf.factor()
        fac_g = ng.factor()
        j = 0
        while j < len(fac_f) and fac_f[j][0] <= Bf:
            j += 1
        smooth_f = (j >= len(fac_f))
        j = 0
        while j < len(fac_g) and fac_g[j][0] <= Bg:
            j += 1
        smooth_g = (j >= len(fac_g))
        if smooth_f and smooth_g:
            if row_id >= rows: # add a row
                mat.append([0 for j in range(cols)])
                if lc_g != 1:
                    mat[row_id][idx_inv_V] = 1

            str_f = ""
            for (l,ei) in fac_f:
                Fl = GF(l)
                Flz.<z> = Fl[]
                ideal_f = gcd(Flz(f), Flz(a-b*x))

                # identify which ones of the ideals of norm l are actually involved in a-b*x
                for id_x in range(len(Ff[l])):
                    poly_id_x = Ff[l][id_x]
                    if gcd(Flz(poly_id_x), ideal_f) > 1:
                        mat[row_id][idx_ideal_f[(l,poly_id_x)]] = -ei ##### HERE -ei for the right-hand-side of the matrix
                        str_f += "("+str(l)+","+str(poly_id_x)+")"
                        if ei>1: str_f+= "^"+str(ei)
            for (l,ei) in fac_g:
                mat[row_id][idx_ideal_g[l]] = ei
                
            print("{:3},{:2}&{:4}& {:15s}& {:4}& {:15s}&{:31s}\\\\".format(a,b, ng,pretty_print_tex_product(ng,fac_g), nf,pretty_print_tex_product(nf,fac_f), str_f))
            row_id += 1
print("")

if row_id >= cols:
    M = Matrix(ZZ, row_id, cols, mat) # over ZZ for nice printing
    print M
    latex.matrix_delimiters(left='[', right=']')
    print latex(M) # see http://doc.sagemath.org/html/en/tutorial/latex.html
    Mr = Matrix(Fr, row_id, cols, mat) # mod r for solving
    print Mr.rank()
    print Mr.right_kernel()
else:
    print("not enough relations, found {}, required {}".format(row_id, cols))


"""
p = 1109 is prime
(p-1)/4 = 277 is prime
Polynomial selection: complex quadratic
m = 571 # root of f and g mod p
f = x^2 + 5
g = 2*x - 33
f = (z + 538) * (z + 571) mod p
Cl(Kf) = 2
units(Kf) = ()
number of real roots, pairs of complex roots: (0, 1)
resultant(f,g) = 1109 = 0 mod p
disc(f) = -20 = -1 * 2^2 * 5

chosen smoothness bounds: f-side 13, g-side 13
compute the factor basis on f-side: factor f mod l for each prime 2 <= l <= Bf=13
2: (z + 1)^2
3: (z + 1) * (z + 2)
5: z^2
7: (z + 3) * (z + 4)
11: z^2 + 5
13: z^2 + 5

factor basis g: [2, 3, 5, 7, 11, 13]
Class number(Kf) = 2
ideals above primes on f-side:
( 2,x + 1 ), generator: no
( 3,x + 1 ), generator: no
( 3,x + 2 ), generator: no
( 5,x     ), generator -af
( 7,x + 3 ), generator: no
( 7,x + 4 ), generator: no

the correspondance between ideals and column indices is the folowing
{2: 0, 3: 1, 5: 2, 7: 3, 11: 4, 13: 5}
{(2, x + 1): 7, (7, x + 3): 11, (7, x + 4): 12, (5, x): 10, (3, x + 1): 8, (3, x + 2): 9}
13 columns, 13 relations meeded
start relation collection, a in [-11,11], b in [1,11]

-11, 1& -55& -5\cd 11       &  126& 2\cd 3^2\cd 7  &(2,x + 1)(3,x + 2)^2(7,x + 4)  \\
-11, 8&-286& -2\cd 11\cd 13 &  441& 3^2\cd 7^2     &(3,x + 1)^2(7,x + 4)^2         \\
 -3, 1& -39& -3\cd 13       &   14& 2\cd 7         &(2,x + 1)(7,x + 3)             \\
 -1, 1& -35& -5\cd 7        &    6& 2\cd 3         &(2,x + 1)(3,x + 1)             \\
  0, 1& -33& -3\cd 11       &    5& 5              &(5,x)                          \\
  1, 2& -64& -2^6           &   21& 3\cd 7         &(3,x + 1)(7,x + 3)             \\
  1, 4&-130& -2\cd 5\cd 13  &   81& 3^4            &(3,x + 2)^4                    \\
  3, 1& -27& -3^3           &   14& 2\cd 7         &(2,x + 1)(7,x + 4)             \\
  4, 1& -25& -5^2           &   21& 3\cd 7         &(3,x + 2)(7,x + 3)             \\
  5, 2& -56& -2^3\cd 7      &   45& 3^2\cd 5       &(3,x + 2)^2(5,x)               \\
 10, 1& -13& -13            &  105& 3\cd 5\cd 7    &(3,x + 2)(5,x)(7,x + 4)        \\
 11, 1& -11& -11            &  126& 2\cd 3^2\cd 7  &(2,x + 1)(3,x + 1)^2(7,x + 3)  \\
 11, 8&-242& -2\cd 11^2     &  441& 3^2\cd 7^2     &(3,x + 2)^2(7,x + 3)^2         \\

\left(\begin{array}{rrrrrrrrrrrrr}
0 & 0 & 1 & 0 & 1 & 0 & 1 & -1 &  0 & -2 &  0 &  0 & -1 \\
1 & 0 & 0 & 0 & 1 & 1 & 1 &  0 & -2 &  0 &  0 &  0 & -2 \\
0 & 1 & 0 & 0 & 0 & 1 & 1 & -1 &  0 &  0 &  0 & -1 &  0 \\
0 & 0 & 1 & 1 & 0 & 0 & 1 & -1 & -1 &  0 &  0 &  0 &  0 \\
0 & 1 & 0 & 0 & 1 & 0 & 1 &  0 &  0 &  0 & -1 &  0 &  0 \\
6 & 0 & 0 & 0 & 0 & 0 & 1 &  0 & -1 &  0 &  0 & -1 &  0 \\
1 & 0 & 1 & 0 & 0 & 1 & 1 &  0 &  0 & -4 &  0 &  0 &  0 \\
0 & 3 & 0 & 0 & 0 & 0 & 1 & -1 &  0 &  0 &  0 &  0 & -1 \\
0 & 0 & 2 & 0 & 0 & 0 & 1 &  0 &  0 & -1 &  0 & -1 &  0 \\
3 & 0 & 0 & 1 & 0 & 0 & 1 &  0 &  0 & -2 & -1 &  0 &  0 \\
0 & 0 & 0 & 0 & 0 & 1 & 1 &  0 &  0 & -1 & -1 &  0 & -1 \\
0 & 0 & 0 & 0 & 1 & 0 & 1 & -1 & -2 &  0 &  0 & -1 &  0 \\
1 & 0 & 0 & 0 & 2 & 0 & 1 &  0 &  0 & -2 &  0 & -2 &  0
\end{array}\right)

12
Vector space of degree 13 and dimension 1 over Finite Field of size 277
Basis matrix:
[  1 219  40  34  79 269 276 139 211   8  20  71 240]
#if we forget to put - in the exponents of the rhs we get this:
#[  1 219  40  34  79 269 276 138  66 269 257 206  37] 

"""                    
