"""
Author Aurore Guillevic
email  aurore.guillevic@inria.fr
date   May 3, 2019

Example of NFS with base-m polynomial selection for p=1109.
"""
# import printing functions for latex output
from examples_pretty_print import pretty_print_product, pretty_print_tex_product, pretty_print_units

p = 1109     # prime, p = 1 mod 4
r = (p-1)//4 # prime order of the subgroup
Fr = GF(r)   # for the linear algebra
Fp = GF(p)

ZZx.<x> = ZZ[]

print("p = {} is prime".format(p))
print("(p-1)/4 = {} is prime".format(r))

print("Polynomial selection base-m")

d = 3 # degree of the polynomial f(x)
m = round(p^(1/d)) # m will be a root of f(x) mod p
f = ZZx(p.digits(m)) # so that f(m) = p
# you may try f = x^3 + x^2 + 9 and f = x^3 + x^2 + x - 1
# f = x^3+x^2+x-1
g = x-m

assert f.is_irreducible()
assert (f(m) % p) == 0
assert (f.resultant(g) % p) == 0

print("m = {} # root of f and g mod p\nf = {}\ng = {}".format(m,f,g))

Fpz.<z> = Fp[] # polynomial ring modulo p
fp = Fpz(f)    # polynomial f modulo p
print("f = {} mod p".format(fp.factor()))
Kf.<af> = NumberField(f)
Clf = Kf.class_number()
print("Cl(Kf) = {} # class number".format(Clf))
print("units(Kf) = {}".format(Kf.units()))
print("number of real roots, pairs of complex roots: {}".format(Kf.signature()))

print("resultant(f,g) = {} = {} mod p".format(f.resultant(g), f.resultant(g) % p))
disc_f =  f.discriminant()
print("disc(f) = {} = {}\n".format(disc_f, disc_f.factor()))

A = 11 # a,b in the relation collection step: a in [-A,A], b in [1,A]

Bf=13
Bg=13
print("chosen smoothness bounds: f-side {}, g-side {}".format(Bf,Bg))


# a dictionary for the factor basis on each side
Ff = dict()
Fg = dict()
print("compute the factor basis on f-side: factor f mod l for each prime 2 <= l <= Bf={}".format(Bf))
for l in prime_range(2, Bf+1):
    Fl = GF(l)     # prime finite field of cardinality l
    Flz.<z> = Fl[] # polynomial ring over Fl
    fl = Flz(f)    # f mod l
    if len(fl.roots())>0: # f has root(s) mod l, so there are as many distinct degree 1 prime ideals above l as distinct roots
        Ff[l] = [ZZx(fi) for (fi,ei) in fl.factor() if fi.degree() == 1]
    print("{}: {}".format(l, (fl).factor()))

Fg = [l for l in prime_range(Bg+1)]
print("\nfactor basis g: {}".format(Fg))

F_f = dict() # a second dict with generators of the ideals
for (l,li) in Ff.items():
    for d in li:
        Il = Kf.ideal([l,d])
        gen_Il = Il.gens_reduced()
        if len(gen_Il) == 1: # means it is principal and one generator was found
            if F_f.has_key(l):
                F_f[l].append(gen_Il[0])
            else:
                F_f[l] = [gen_Il[0]]

if f == x^3+x^2+9: # adjust this ideal that is actually of degree 2: this is not (3,x) but (3,x^2)
    if Ff[3][1] == x:
        Ff[3][1] = x^2
print("ideals above primes on f-side:")
for l in prime_range(Bf+1):
    if Ff.has_key(l):
        for idx in range(len(Ff[l])):
            print("({:2d},{:6}), generator {}".format(l, Ff[l][idx], F_f[l][idx]))

# Problem: Sage chooses generators of the form af^2+af+1
# http://doc.sagemath.org/html/en/reference/number_fields/sage/rings/number_field/unit_group.html
print "\nUnit group of Kf"
UKf = Kf.unit_group()
u0_f,u1_f = UKf.gens()
torsion_unit_f = Kf(u0_f)
fond_unit_f = Kf(u1_f)
print("the generators of the unit group are ({},{}), in Kf: ({},{})".format(u0_f, u1_f, torsion_unit_f, fond_unit_f))
print("the generators of the unit group are of order: {}".format([ui_f.multiplicative_order() for ui_f in UKf.gens()]))
print("the unit group has rank {}".format(UKf.rank()))

# define two dictionaries to map each prime of Fg, and each ideal of Ff, to a column index in the matrix,
# in order to translate each relation into a row in the matrix
# since g is rational, put Fg first in the matrix, starting with 2, so that the logarithms will be in basis 2
idx_ideal_g = {Fg[idx]:idx for idx in range(len(Fg))}
idx = len(Fg)
idx_unit_f = idx
idx_ideal_f = dict()
idx_ideal_f[(1,fond_unit_f.polynomial())] = idx_unit_f
idx += 1
#for (l,li) in Ff.items(): it does not provide the ideals by increasing size, so do it explicitly
for l in prime_range(Bf+1):
    if Ff.has_key(l):
        for id_li in Ff[l]:
            idx_ideal_f[(l,id_li)] = idx
            idx += 1

print("\nthe correspondance between ideals and column indices is the folowing")
print idx_ideal_g
print idx_ideal_f

row_id = 0
rows = idx
cols = idx
print("{} columns, {} relations meeded".format(cols,rows))

# here both polynomials are monic so there is no equivalent of (1/V)
mat = [[0 for j in range(cols)] for i in range(rows)]

print("start relation collection, a in [-{0},{0}], b in [1,{0}]".format(A))
for a in range(-A,A+1):
    for b in [b for b in range(1,A+1) if gcd(a,b) == 1]:
        nf = ZZ(f.resultant(a-b*x))
        ng = a-b*m
        if nf == 0 or ng == 0:
            continue
        fac_f = nf.factor()
        fac_g = ng.factor()
        j = 0
        while j < len(fac_f) and fac_f[j][0] <= Bf:
            j += 1
        smooth_f = (j >= len(fac_f))
        j = 0
        while j < len(fac_g) and fac_g[j][0] <= Bg:
            j += 1
        smooth_g = (j >= len(fac_g))
        if smooth_f and smooth_g:
            if row_id >= rows: # add a row
                mat.append([0 for j in range(cols)])
            str_f = ""
            # prod_id_f will be the product of generators of ideals,
            # so that a-b*af = prod_id_f up to a unit to compute: u = (a-b*af) / prod_id_f
            prod_id_f = Kf(1)
            for (l,ei) in fac_f: # l a prime, ei an exponent: l^ei | fac_f
                # to which ideal above l does the factor (l^ei) correspond to?
                Fl = GF(l)
                Flz.<z> = Fl[]
                ideal_f = gcd(Flz(f), Flz(a-b*x))
                for id_x in range(len(F_f[l])): # identify which ones of the ideals of norm l are actually involved in a-b*x
                    # pb: if id_x.polynomial() contains 1/l, then Flz(1/l) fails.
                    ideal_id_x = F_f[l][id_x]
                    poly_id_x = Ff[l][id_x]
                    if gcd(Flz(poly_id_x), ideal_f) > 1:
                        if l == 3 and ideal_f == Flz(x) and f == x^3+x^2+9: # special case, this ideal is of degree 2
                            prod_id_f *= ideal_id_x^(ei//2)
                            mat[row_id][idx_ideal_f[(l,poly_id_x)]] = -ei//2
                            str_f += "("+str(l)+","+str(poly_id_x)+")"
                            if (ei//2)>1: str_f+= "^"+str(ei//2)
                        else:
                            prod_id_f *= ideal_id_x^ei
                            mat[row_id][idx_ideal_f[(l,poly_id_x)]] = -ei
                            str_f += "("+str(l)+","+str(poly_id_x)+")"
                            if ei>1: str_f+= "^"+str(ei)
            unf = (a-b*af)/prod_id_f # unit
            if not abs(unf.norm()) == 1:
                print("error in computing (a-b*af)/(prod(generator ideals))")
            # compute the contribution of the units: u0=-1 and the fundamental unit u1
            (log_u0, log_u1) = UKf.log(unf)
            # this should be: (-1)^log_u0*(uf)^log_u1 == unf
            if log_u0 == 0: unfs=" "
            else: unfs="-"            
            mat[row_id][idx_unit_f] = -log_u1 # exponent
            str_uf = pretty_print_units(log_u1, unfs, "f")

            # the rational side is much easier: no unit, no ideal, only prime integers
            for (l,ei) in fac_g:
                mat[row_id][idx_ideal_g[l]] = ei

            print("{:3},{:2}&{:4}& {:11s}& {:4}& {:15s}&{:31s}&{}\\\\".format(a,b, ng,pretty_print_tex_product(ng,fac_g), nf,pretty_print_tex_product(nf,fac_f), str_f, str_uf))
            row_id += 1
print("")

if row_id >= cols:
    M = Matrix(ZZ, row_id, cols, mat)
    print M
    latex.matrix_delimiters(left='[', right=']')
    print latex(M) # see http://doc.sagemath.org/html/en/tutorial/latex.html
    Mr = Matrix(Fr, row_id, cols, mat)
    print("M mod r has rank {}".format(Mr.rank()))
    print Mr.right_kernel()
else:
    print("not enough relations, found {}, required {}".format(row_id, cols))
